@@ -5,19 +5,20 @@
 
 // FIXME: this badly needs rename/rewrite (matklad, 2020-02-06).
 
+use arrayvec::ArrayVec;
 use hir::{
     Field, GenericParam, HasVisibility, Impl, Label, Local, MacroDef, Module, ModuleDef, Name,
     PathResolution, Semantics, Visibility,
 };
 use syntax::{
-    ast::{self, AstNode, PathSegmentKind},
-    match_ast, SyntaxKind, SyntaxNode,
+    ast::{self, AstNode},
+    match_ast, SyntaxKind, SyntaxNode, SyntaxToken,
 };
 
-use crate::RootDatabase;
+use crate::{helpers::try_resolve_derive_input_at, RootDatabase};
 
 // FIXME: a more precise name would probably be `Symbol`?
-#[derive(Debug, PartialEq, Eq, Copy, Clone)]
+#[derive(Debug, PartialEq, Eq, Copy, Clone, Hash)]
 pub enum Definition {
     Macro(MacroDef),
     Field(Field),
@@ -29,6 +30,73 @@ pub enum Definition {
 }
 
 impl Definition {
+    pub fn from_token(
+        sema: &Semantics<RootDatabase>,
+        token: &SyntaxToken,
+    ) -> ArrayVec<Definition, 2> {
+        let parent = match token.parent() {
+            Some(parent) => parent,
+            None => return Default::default(),
+        };
+        let attr = parent
+            .ancestors()
+            .find_map(ast::TokenTree::cast)
+            .and_then(|tt| tt.parent_meta())
+            .and_then(|meta| meta.parent_attr());
+        if let Some(attr) = attr {
+            try_resolve_derive_input_at(&sema, &attr, &token)
+                .map(Definition::Macro)
+                .into_iter()
+                .collect()
+        } else {
+            Self::from_node(sema, &parent)
+        }
+    }
+
+    pub fn from_node(sema: &Semantics<RootDatabase>, node: &SyntaxNode) -> ArrayVec<Definition, 2> {
+        let mut res = ArrayVec::new();
+        (|| {
+            match_ast! {
+                match node {
+                    ast::Name(name) => {
+                        match NameClass::classify(&sema, &name)? {
+                            NameClass::Definition(it) | NameClass::ConstReference(it) => res.push(it),
+                            NameClass::PatFieldShorthand { local_def, field_ref } => {
+                                res.push(Definition::Local(local_def));
+                                res.push(Definition::Field(field_ref));
+                            }
+                        }
+                    },
+                    ast::NameRef(name_ref) => {
+                        match NameRefClass::classify(sema, &name_ref)? {
+                            NameRefClass::Definition(it) => res.push(it),
+                            NameRefClass::FieldShorthand { local_ref, field_ref } => {
+                                res.push(Definition::Local(local_ref));
+                                res.push(Definition::Field(field_ref));
+                            }
+                        }
+                    },
+                    ast::Lifetime(lifetime) => {
+                        let def = if let Some(x) = NameClass::classify_lifetime(&sema, &lifetime) {
+                            NameClass::defined(x)
+                        } else {
+                            NameRefClass::classify_lifetime(&sema, &lifetime).and_then(|class| match class {
+                                NameRefClass::Definition(it) => Some(it),
+                                _ => None,
+                            })
+                        };
+                        if let Some(def) = def {
+                            res.push(def);
+                        }
+                    },
+                    _ => (),
+                }
+            }
+            Some(())
+        })();
+        res
+    }
+
     pub fn module(&self, db: &RootDatabase) -> Option<Module> {
         match self {
             Definition::Macro(it) => it.module(db),
@@ -44,19 +112,7 @@ impl Definition {
     pub fn visibility(&self, db: &RootDatabase) -> Option<Visibility> {
         match self {
             Definition::Field(sf) => Some(sf.visibility(db)),
-            Definition::ModuleDef(def) => match def {
-                ModuleDef::Module(it) => Some(it.visibility(db)),
-                ModuleDef::Function(it) => Some(it.visibility(db)),
-                ModuleDef::Adt(it) => Some(it.visibility(db)),
-                ModuleDef::Const(it) => Some(it.visibility(db)),
-                ModuleDef::Static(it) => Some(it.visibility(db)),
-                ModuleDef::Trait(it) => Some(it.visibility(db)),
-                ModuleDef::TypeAlias(it) => Some(it.visibility(db)),
-                // NB: Variants don't have their own visibility, and just inherit
-                // one from the parent. Not sure if that's the right thing to do.
-                ModuleDef::Variant(it) => Some(it.parent_enum(db).visibility(db)),
-                ModuleDef::BuiltinType(_) => None,
-            },
+            Definition::ModuleDef(def) => Some(def.visibility(db)),
             Definition::Macro(_)
             | Definition::SelfType(_)
             | Definition::Local(_)
@@ -146,29 +202,23 @@ impl NameClass {
                     if let Some(use_tree) = it.syntax().parent().and_then(ast::UseTree::cast) {
                         let path = use_tree.path()?;
                         let path_segment = path.segment()?;
-                        let name_ref_class = path_segment
-                            .kind()
-                            .and_then(|kind| {
-                                match kind {
-                                    // The rename might be from a `self` token, so fallback to the name higher
-                                    // in the use tree.
-                                    PathSegmentKind::SelfKw => {
-                                        let use_tree = use_tree
-                                            .syntax()
-                                            .parent()
-                                            .as_ref()
-                                            // Skip over UseTreeList
-                                            .and_then(SyntaxNode::parent)
-                                            .and_then(ast::UseTree::cast)?;
-                                        let path = use_tree.path()?;
-                                        let path_segment = path.segment()?;
-                                        path_segment.name_ref()
-                                    },
-                                    PathSegmentKind::Name(name_ref) => Some(name_ref),
-                                    _ => return None,
-                                }
-                            })
-                            .and_then(|name_ref| NameRefClass::classify(sema, &name_ref))?;
+                        let name_ref = path_segment.name_ref()?;
+                        let name_ref = if name_ref.self_token().is_some() {
+                             use_tree
+                                .syntax()
+                                .parent()
+                                .as_ref()
+                                // Skip over UseTreeList
+                                .and_then(|it| {
+                                    let use_tree = it.parent().and_then(ast::UseTree::cast)?;
+                                    let path = use_tree.path()?;
+                                    let path_segment = path.segment()?;
+                                    path_segment.name_ref()
+                                }).unwrap_or(name_ref)
+                        } else {
+                            name_ref
+                        };
+                        let name_ref_class = NameRefClass::classify(sema, &name_ref)?;
 
                         Some(NameClass::Definition(match name_ref_class {
                             NameRefClass::Definition(def) => def,
@@ -353,7 +403,7 @@ impl NameRefClass {
                             hir::AssocItem::TypeAlias(it) => Some(*it),
                             _ => None,
                         })
-                        .find(|alias| &alias.name(sema.db).to_string() == &name_ref.text())
+                        .find(|alias| alias.name(sema.db).to_string() == name_ref.text())
                     {
                         return Some(NameRefClass::Definition(Definition::ModuleDef(
                             ModuleDef::TypeAlias(ty),
